import emerge as em
import numpy as np
from emerge.plot import plot_sp

""" DEMO: COMBLINE FILTER

In this demo we will look at the design of a combline filter in EMerge. The filter design was taken from
the book "Modern RF and Microwave Filter Design" by Protap Pramanick and Prakash Bhartia.
Some of the dimensions where not clear.

In this demo we will look at the Modeler class and how it can help us quickly create complicated geometries.

"""

# First we define some quantities for our simulation.
mm = 0.001
mil = 0.0254*mm

a = 240*mil
b = 248*mil
d1 = 10*mil
d2 = 10*mil
dc = 8.5*mil
lr1 = b-d1
lr2 = b-d2
W = 84*mil
S1 = 117*mil
S2 = 136*mil
C1 = b-dc
h = 74*mil
wi = 84*mil
Lbox = 5*W + 2*(S1+S2+wi)

x1 = wi+W/2
x2 = x1 + W + S1
x3 = x2 + W + S2
x4 = x3 + W + S2
x5 = x4 + W + S1

rout = 40.5*mil
rin = 12.5*mil
lfeed = 100*mil

# A usual we start our simulation file
model = em.Simulation('Combline_DEMO', loglevel='DEBUG')


# The filter consists of quarter lamba cylindrical pins inside an airbox.
# First we create the airbox
box = em.geo.Box(Lbox, a, b, position=(0,-a/2,0))

# Next we create 5 cyllinders using the Modeler class.
# The modeler class also implements a method chaining interface. In this example we stick to simpler features.
# The modeler class allows us to create a parameter series using the modeler.series() method. We provid it with quantities.
# We can do this for multiple at the same time (as you can also see with the position). The modeler class
# will recognize the multiple quantities and simply create 5 different cyllinders, one for each parameter pair.
stubs = model.modeler.cyllinder(W/2, model.modeler.series(C1, lr1, lr2, lr1, C1), position=(model.modeler.series(x1, x2, x3, x4, x5), 0, 0), NPoly=10)

# Next we create the in and output feed cyllinders for the coaxial cable. We will use the Nsections feature in order to guarantee a better
# adherence to the boundary.
feed1out = em.geo.Cylinder(rout, lfeed, em.CoordinateSystem(em.ZAX, em.YAX, em.XAX, np.array([-lfeed, 0, h])), Nsections=12)
feed1in = em.geo.Cylinder(rin, lfeed+wi+W/2, em.CoordinateSystem(em.ZAX, em.YAX, em.XAX, np.array([-lfeed, 0, h])), Nsections=8)
feed2out = em.geo.Cylinder(rout, lfeed, em.CoordinateSystem(em.ZAX, em.YAX, em.XAX, np.array([Lbox, 0, h])), Nsections=12)
feed2in = em.geo.Cylinder(rin, lfeed+wi+W/2, em.CoordinateSystem(em.ZAX, em.YAX, em.XAX, np.array([Lbox-wi-W/2, 0, h])), Nsections=8)

# Next we subtract the stubs and the center conductor from the box and feedline.
for ro in stubs:
    box = em.geo.subtract(box, ro)
box = em.geo.subtract(box, feed1in, remove_tool=False)
box = em.geo.subtract(box, feed2in, remove_tool=False)
feed1out = em.geo.subtract(feed1out, feed1in)
feed2out = em.geo.subtract(feed2out, feed2in)

# Finally we may define our geometry
model.commit_geometry()

model.view()

# We define our frequency range and a fine sampling.
model.mw.set_frequency_range(6e9, 8e9, 21)

model.mw.resolution = 0.05

# To improve simulation quality we refine the faces at the top of the cylinders.
for stub in stubs:
    model.mesher.set_boundary_size(box.face('back', tool=stub), 0.0002)

# Finally we may create our mesh.
model.generate_mesh()

# We define our modal ports, assign the boundary condition and execute a modal analysis to solve for the
# coaxial field mode.
port1 = model.mw.bc.ModalPort(model.select.face.near(-lfeed, 0, h), 1, TEM=True)
port2 = model.mw.bc.ModalPort(model.select.face.near(Lbox+lfeed, 0, h), 2, TEM=True)

# At last we can compute the frequency domain study
data = model.mw.run_sweep(parallel=True)

# Next we will use the Vector Fitting algorithm to model our S-parameters with a Rational function

fdense = np.linspace(6e9, 9e9, 2001)

S11 = data.scalar.grid.model_S(1,1,fdense)
S21 = data.scalar.grid.model_S(2,1,fdense)

plot_sp(fdense/1e9, [S11, S21], labels=['S11','S21'])

# We can also plot the field inside. First we create a grid of sample point coordinates
xs = np.linspace(0, Lbox, 41)
ys = np.linspace(-a/2, a/2, 11)
zs = np.linspace(0, b, 15)

X, Y, Z = np.meshgrid(xs, ys, zs, indexing='ij')
X = X.flatten()
Y = Y.flatten()
Z = Z.flatten()

# The E-field can be interpolated by selecting a desired solution and then interpolating it.
Ex, Ey, Ez = data.field[3].interpolate(X,Y,Z).E

# We can add the objects we want and fields using the shown methods.
model.display.add_object(box, opacity=0.1, show_edges=True)
model.display.add_quiver(X,Y,Z, Ex.real, Ey.real, Ez.real)
model.display.add_object(feed1out, opacity=0.1)
model.display.add_portmode(port1, 21)
model.display.add_portmode(port2, 21)
model.display.add_surf(*data.field[3].cutplane(ds=0.5*mm, y=0).scalar('normE'))
model.display.show()


# EMerge is an open source Python based FEM EM simulation module.
# Copyright (C) 2025  Robert Fennis.

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, see
# <https://www.gnu.org/licenses/>.

from __future__ import annotations
from .mesher import Mesher, unpack_lists
from .geometry import GeoObject, _GEOMANAGER
from .geo.modeler import Modeler
from .physics.microwave.microwave_3d import Microwave3D
from .mesh3d import Mesh3D
from .selection import Selector, FaceSelection, Selection
from .logsettings import LOG_CONTROLLER
from .plot.pyvista import PVDisplay
from .dataset import SimulationDataset
from .periodic import PeriodicCell
from .bc import BoundaryCondition
from typing import Literal, Generator, Any
from loguru import logger
import numpy as np
import gmsh # type: ignore
import joblib # type: ignore
import os
import inspect
from pathlib import Path
from atexit import register
import signal


############################################################
#                   EXCEPTION DEFINITIONS                  #
############################################################

_GMSH_ERROR_TEXT = """
--------------------------
Known problems/solutions:
(1) - PLC Error:  A segment and a facet intersect at point
    This can be caused when approximating thin curved volumes. Try to decrease the mesh size for that region.
--------------------------
"""

class SimulationError(Exception):
    pass

############################################################
#                 BASE 3D SIMULATION MODEL                 #
############################################################

class Simulation:

    def __init__(self, 
                 modelname: str, 
                 loglevel: Literal['TRACE','DEBUG','INFO','WARNING','ERROR'] = 'INFO',
                 load_file: bool = False,
                 save_file: bool = False,
                 logfile: bool = False,
                 path_suffix: str = ".EMResults"):
        """Generate a Simulation class object.

        As a minimum a file name should be provided. Additionally you may provide it with any
        class that inherits from BaseDisplay. This will then be used for geometry displaying.

        Args:
            modelname (str): The name of the simulation model. This will be used for filenames and path names when saving data.
            loglevel ("DEBUG","INFO","WARNING","ERROR", optional): The loglevel to use for loguru. Defaults to 'INFO'.
            load_file (bool, optional): If the simulatio model should be loaded from a file. Defaults to False.
            save_file (bool, optional): if the simulation file should be stored to a file. Defaults to False.
            logfile (bool, optional): If a file should be created that contains the entire log of the simulation. Defaults to False.
            path_suffix (str, optional): The suffix that will be added to the results directory. Defaults to ".EMResults".
        """

        caller_file = Path(inspect.stack()[1].filename).resolve()
        base_path = caller_file.parent

        self.modelname = modelname
        self.modelpath = base_path / (modelname.lower()+path_suffix)
        self.mesher: Mesher = Mesher()
        self.modeler: Modeler = Modeler()
        
        self.mesh: Mesh3D = Mesh3D(self.mesher)
        self.select: Selector = Selector()

        ## STATES
        self.__active: bool = False
        self._defined_geometries: bool = False
        self._cell: PeriodicCell | None = None

        self.display: PVDisplay = PVDisplay(self.mesh)

        self.save_file: bool = save_file
        self.load_file: bool = load_file

        self.data: SimulationDataset = SimulationDataset()

        ## Physics
        self.mw: Microwave3D = Microwave3D(self.mesher, self.data.mw)

        self._initialize_simulation()

        self.set_loglevel(loglevel)
        if logfile:
            self.set_logfile()

        self._update_data()
    

    ############################################################
    #                       PRIVATE FUNCTIONS                  #
    ############################################################

    def __setitem__(self, name: str, value: Any) -> None:
        """Store data in the current data container"""
        self.data.sim[name] = value

    def __getitem__(self, name: str) -> Any:
        """Get the data from the current data container"""
        return self.data.sim[name]
    
    def __enter__(self) -> Simulation:
        """This method is depricated with the new atexit system. It still exists for backwards compatibility.

        Returns:
            Simulation: the Simulation object
        """
        return self

    def __exit__(self, type, value, tb):
        """This method no longer does something. It only serves as backwards compatibility."""
        self._exit_gmsh()
        return False
    
    def _install_signal_handlers(self):
        # on SIGINT (Ctrl-C) or SIGTERM, call our exit routine
        for sig in (signal.SIGINT, signal.SIGTERM):
            signal.signal(sig, self._handle_signal)
    
    def _handle_signal(self, signum, frame):
        """
        Signal handler: do our cleanup, then re-raise
        the default handler so that exit code / traceback
        is as expected.
        """
        try:
            # run your atexit-style cleanup
            self._exit_gmsh()
        except Exception:
            # log but don’t block shutdown
            logger.exception("Error during signal cleanup")
        finally:
            # restore default handler and re‐send the signal
            signal.signal(signum, signal.SIG_DFL)
            os.kill(os.getpid(), signum)

    def _initialize_simulation(self):
        """Initializes the Simulation data and GMSH API with proper shutdown routines.
        """
        _GEOMANAGER.sign_in(self.modelname)
        
        # If GMSH is not yet initialized (Two simulation in a file)
        if gmsh.isInitialized() == 0:
            logger.debug('Initializing GMSH')
            gmsh.initialize()
            
            # Set an exit handler for Ctrl+C cases
            self._install_signal_handlers()

            # Restier the Exit GMSH function on proper program abortion
            register(self._exit_gmsh)

        # Create a new GMSH model or load it
        if not self.load_file:
            gmsh.model.add(self.modelname)
            self.data: SimulationDataset = SimulationDataset()
        else:
            self.load()

        # Set the Simulation state to active
        self.__active = True
        return self

    def _exit_gmsh(self):
        # If the simulation object state is still active (GMSH is running)
        if not self.__active:
            return
        logger.debug('Exiting program')
        # Save the file first
        if self.save_file:
            self.save()
        # Finalize GMSH
        if gmsh.isInitialized():
            gmsh.finalize()

        logger.debug('GMSH Shut down successful')
        # set the state to active
        self.__active = False

    def _update_data(self) -> None:
        """Writes the stored physics data to each phyics class insatnce"""
        self.mw.data = self.data.mw

    def all_geometries(self) -> list[GeoObject]:
        """Returns all geometries stored in the simulation file."""
        return [obj for obj in self.data.sim.default.values() if isinstance(obj, GeoObject)]
    
    def all_bcs(self) -> list[BoundaryCondition]:
        """Returns all boundary condition objects stored in the simulation file"""
        return [obj for obj in self.data.sim.default.values() if isinstance(obj, BoundaryCondition)]
    
    def _set_mesh(self, mesh: Mesh3D) -> None:
        """Set the current model mesh to a given mesh."""
        self.mesh = mesh
        self.mw.mesh = mesh
        self.display._mesh = mesh
    
    ############################################################
    #                       PUBLIC FUNCTIONS                  #
    ############################################################

    def save(self) -> None:
        """Saves the current model in the provided project directory."""
        # Ensure directory exists
        if not self.modelpath.exists():
            self.modelpath.mkdir(parents=True, exist_ok=True)
            logger.info(f"Created directory: {self.modelpath}")

        # Save mesh
        mesh_path = self.modelpath / 'mesh.msh'
        brep_path = self.modelpath / 'model.brep'

        gmsh.option.setNumber('Mesh.SaveParametric', 1)
        gmsh.option.setNumber('Mesh.SaveAll', 1)
        gmsh.model.geo.synchronize()
        gmsh.model.occ.synchronize()

        gmsh.write(str(mesh_path))
        gmsh.write(str(brep_path))
        logger.info(f"Saved mesh to: {mesh_path}")

        # Pack and save data
        dataset = dict(simdata=self.data, mesh=self.mesh)
        data_path = self.modelpath / 'simdata.emerge'
        joblib.dump(dataset, str(data_path))
        logger.info(f"Saved simulation data to: {data_path}")

    def load(self) -> None:
        """Loads the model from the project directory."""
        mesh_path = self.modelpath / 'mesh.msh'
        brep_path = self.modelpath / 'model.brep'
        data_path = self.modelpath / 'simdata.emerge'

        if not mesh_path.exists() or not data_path.exists():
            raise FileNotFoundError("Missing required mesh or data file.")

        # Load mesh
        gmsh.open(str(brep_path))
        gmsh.merge(str(mesh_path))
        gmsh.model.geo.synchronize()
        gmsh.model.occ.synchronize()
        logger.info(f"Loaded mesh from: {mesh_path}")
        #self.mesh.update([])

        # Load data
        datapack = joblib.load(str(data_path))
        self.data = datapack['simdata']
        self._set_mesh(datapack['mesh'])
        logger.info(f"Loaded simulation data from: {data_path}")
    
    def set_loglevel(self, loglevel: Literal['DEBUG','INFO','WARNING','ERROR']) -> None:
        """Set the loglevel for loguru.

        Args:
            loglevel ('DEBUG','INFO','WARNING','ERROR'): The loglevel
        """
        LOG_CONTROLLER.set_std_loglevel(loglevel)
        if loglevel not in ('TRACE','DEBUG'):
            gmsh.option.setNumber("General.Terminal", 0)

    def set_logfile(self) -> None:
        """Adds a file output for the logger."""
        LOG_CONTROLLER.set_write_file(self.modelpath)
        
    def view(self, 
             selections: list[Selection] | None = None, 
             use_gmsh: bool = False,
             volume_opacity: float = 0.1,
             surface_opacity: float = 1,
             show_edges: bool = True) -> None:
        """View the current geometry in either the BaseDisplay object (PVDisplay only) or
        the GMSH viewer.

        Args:
            selections (list[Selection], optional): Additional selections to highlight. Defaults to None.
            use_gmsh (bool, optional): Whether to use the GMSH display. Defaults to False.
            opacity (float, optional): The global opacity of all objects.. Defaults to None.
            show_edges (bool, optional): Whether to show the geometry edges. Defaults to None.
        """
        if not (self.display is not None and self.mesh.defined) or use_gmsh:
            gmsh.model.occ.synchronize()
            gmsh.fltk.run()
            return
        for geo in _GEOMANAGER.all_geometries():
            self.display.add_object(geo)
        if selections:
            [self.display.add_object(sel, color='red', opacity=0.3) for sel in selections]
        self.display.show()

        return None
        
    def set_periodic_cell(self, cell: PeriodicCell, excluded_faces: list[FaceSelection] | None = None):
        """Set the given periodic cell object as the simulations peridicity.

        Args:
            cell (PeriodicCell): The PeriodicCell class
            excluded_faces (list[FaceSelection], optional): Faces to exclude from the periodic boundary condition. Defaults to None.
        """
        self.mw.bc._cell = cell
        self._cell = cell

    def commit_geometry(self, *geometries: GeoObject | list[GeoObject]) -> None:
        """Finalizes and locks the current geometry state of the simulation.

        The geometries may be provided (legacy behavior) but are automatically managed in the background.
        
        """
        geometries_parsed: Any = None
        if not geometries:
            geometries_parsed = _GEOMANAGER.all_geometries()
        else:
            geometries_parsed = unpack_lists(geometries + tuple([item for item in self.data.sim.default.values() if isinstance(item, GeoObject)]))
        
        self.data.sim['geometries'] = geometries_parsed
        self.mesher.submit_objects(geometries_parsed)
        self._defined_geometries = True
        self.display._facetags = [dt[1] for dt in gmsh.model.get_entities(2)]
          
    def generate_mesh(self) -> None:
        """Generate the mesh. 
        This can only be done after commit_geometry(...) is called and if frequencies are defined.

        Args:
            name (str, optional): The mesh file name. Defaults to "meshname.msh".

        Raises:
            ValueError: ValueError if no frequencies are defined.
        """
        if not self._defined_geometries:
            self.commit_geometry()
        
        # Set the cell periodicity in GMSH
        if self._cell is not None:
            self.mesher.set_periodic_cell(self._cell)
            
        self.mw._initialize_bcs()

        # Check if frequencies are defined: TODO: Replace with a more generic check
        if self.mw.frequencies is None:
            raise ValueError('No frequencies defined for the simulation. Please set frequencies before generating the mesh.')

        gmsh.model.occ.synchronize()

        # Set the mesh size
        self.mesher.set_mesh_size(self.mw.get_discretizer(), self.mw.resolution)
        
        try:
            gmsh.logger.start()
            gmsh.model.mesh.generate(3)
            logs = gmsh.logger.get()
            gmsh.logger.stop()
            for log in logs:
                logger.trace('[GMSH] '+log)
        except Exception:
            logger.error('GMSH Mesh error detected.')
            print(_GMSH_ERROR_TEXT)
            raise
        self.mesh.update(self.mesher._get_periodic_bcs())
        self.mesh.exterior_face_tags = self.mesher.domain_boundary_face_tags
        gmsh.model.occ.synchronize()
        self._set_mesh(self.mesh)

    def parameter_sweep(self, clear_mesh: bool = True, **parameters: np.ndarray) -> Generator[tuple[float,...], None, None]:
        """Executes a parameteric sweep iteration.

        The first argument clear_mesh determines if the mesh should be cleared and rebuild in between sweeps. This is usually needed
        except when you change only port-properties or material properties. The parameters of the sweep can be provided as a set of 
        keyword arguments. As an example if you defined the axes: width=np.linspace(...) and height=np.linspace(...). You can
        add them as arguments using .parameter_sweep(True, width=width, height=height).

        The rest of the simulation commands should be inside the iteration scope

        Args:
            clear_mesh (bool, optional): Wether to clear the mesh in between sweeps. Defaults to True.

        Yields:
            Generator[tuple[float,...], None, None]: The parameters provided

        Example:
         >>> for W, H in model.parameter_sweep(True, width=widths, height=heights):
         >>>    // build simulation
         >>>    data = model.run_sweep()
         >>> // Extract the data
         >>> widths, heights, frequencies, S21 = data.ax('width','height','freq').S(2,1)
        """
        paramlist = sorted(list(parameters.keys()))
        dims = np.meshgrid(*[parameters[key] for key in paramlist], indexing='ij')
        dims_flat = [dim.flatten() for dim in dims]
        self.mw.cache_matrices = False
        for i_iter in range(dims_flat[0].shape[0]):
            if clear_mesh:
                logger.info('Cleaning up mesh.')
                gmsh.clear()
                mesh = Mesh3D(self.mesher)
                _GEOMANAGER.reset(self.modelname)
                self._set_mesh(mesh)
                self.mw.reset()
            
            params = {key: dim[i_iter] for key,dim in zip(paramlist, dims_flat)}
            self.mw._params = params
            self.data.sim.new(**params)

            logger.info(f'Iterating: {params}')
            if len(dims_flat)==1:
                yield (dims_flat[0][i_iter],)
            else:
                yield (dim[i_iter] for dim in dims_flat) # type: ignore
        self.mw.cache_matrices = True
    
    def export(self, filename: str):
        """Exports the model or mesh depending on the extension. 
        
        Exporting is realized by GMSH.
        Supported file formats are:
        
        3D Model: .opt, .geo_unrolled, .brep, .xao ,.step and .iges
        Mesh: .msh, .inp, .key, ._0000.rad, .celum, .cgns, .diff, .unv, .ir3, .mes, .mesh 
              .mail, .m, .bdf, .off, .p3d, .stl, .wrl, .vtk, .dat, .ply2, .su2, .neu, .x3d

        Args:
            filename (str): The filename
        """
        gmsh.write(filename)
    ############################################################
    #                     DEPRICATED FUNCTIONS                #
    ############################################################

    def define_geometry(self, *args):
        """DEPRICATED VERSION: Use .commit_geometry()
        """
        logger.warning('define_geometry() will be derpicated. Use commit_geometry() instead.')
        self.commit_geometry(*args)
        


# EMerge is an open source Python based FEM EM simulation module.
# Copyright (C) 2025  Robert Fennis.

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, see
# <https://www.gnu.org/licenses/>.

from __future__ import annotations
import gmsh
import numpy as np
from typing import Literal, Callable
from ..geometry import GeoEdge, GeoSurface, GeoVolume



class Curve(GeoEdge):
    
    
    def __init__(self, xpts: np.ndarray, ypts: np.ndarray, zpts: np.ndarray, 
                 degree: int = 3,
                 weights: list[float] | None = None,
                 knots: list[float] | None = None,
                 ctype: Literal['Spline','BSpline','Bezier'] = 'Bezier'):
        self.xpts: np.ndarray = xpts
        self.ypts: np.ndarray = ypts
        self.zpts: np.ndarray = zpts
        
        points = [gmsh.model.occ.add_point(x,y,z) for x,y,z in zip(xpts, ypts, zpts)]
        
        if ctype.lower()=='spline':
            tags = gmsh.model.occ.addSpline(points)
            
        elif ctype.lower()=='bspline':
            if weights is None:
                weights = []
            if knots is None:
                knots = []
            tags = gmsh.model.occ.addBSpline(points, degree=degree, weights=weights, knots=knots)
        else:
            tags = gmsh.model.occ.addBezier(points)
        
        tags = gmsh.model.occ.addWire([tags,])
        gmsh.model.occ.remove([(0,tag) for tag in points])
        super().__init__(tags)
    



class Helix(GeoVolume):
    
    
        


from typing import Any
from dataclasses import dataclass
import math
import re

def rotation_angle(a: tuple[float, float], b: tuple[float, float]) -> float:
    """
    Returns the signed angle in degrees you must rotate vector a
    to align with vector b. Positive = CCW, Negative = CW.
    
    a, b: (vx, vy) tuples in the plane.
    """
    ax, ay = a
    bx, by = b

    # compute 2D cross product (scalar) and dot product
    cross = ax * by - ay * bx
    dot   = ax * bx + ay * by

    # atan2(cross, dot) gives angle between -π and π
    angle_rad = math.atan2(cross, dot)
    angle_deg = math.degrees(angle_rad)
    return -angle_deg

@dataclass
class Instruction:
    instr: str
    args: tuple[int | float, ...]
    kwargs: dict[str,float] | None = None

    def __post_init__(self):
        if self.kwargs is None:
            self.kwargs = {}

symbols = ['=','>','v','^','<','@','/','\\','T']
char_class = ''.join(re.escape(c) for c in symbols)

pattern = re.compile(rf'([{char_class}])([\d\,\.\-]+)')

def parse_macro(pathstring: str, width: int | float, direction: tuple[float, float]) -> list[Instruction]:
    instructions = pattern.findall(pathstring.replace(' ',''))
    
    oi = []
    for com, val in instructions:
        if ',' in val:
            val_list: list[float] = [float(x) for x in val.split(',')]
            ival, width = val_list[0], val_list[1]
        else:
            ival = float(val)
        if com == '=':
            oi.append(Instruction('straight',(ival,),{'width': width})) #type: ignore
        elif com == '>':
            oi.append(Instruction('turn',(rotation_angle(direction, (1., 0.)),) ))
            oi.append(Instruction('straight',(ival,),{'width': width}))
            direction = (1. ,0. )
        elif com == '<':
            oi.append(Instruction('turn',(rotation_angle(direction, (-1., 0.)),) ))
            oi.append(Instruction('straight',(ival,),{'width': width}))
            direction = (-1.,0.)
        elif com == 'v':
            oi.append(Instruction('turn',(rotation_angle(direction, (0.,-1.)),) ))
            oi.append(Instruction('straight',(ival,),{'width': width}))
            direction = (0.,-1.)
        elif com == '^':
            oi.append(Instruction('turn',(rotation_angle(direction, (0.,1.)),) ))
            oi.append(Instruction('straight',(ival,),{'width': width}))
            direction = (0.,1.)
        elif com == '\\':
            oi.append(Instruction('turn',(90,) ))
            oi.append(Instruction('straight',(ival,),{'width': width}))
            direction = (direction[1],-direction[0])
        elif com == '/':
            oi.append(Instruction('turn',(-90,) ))
            oi.append(Instruction('straight',(ival,),{'width': width}))
            direction = (-direction[1],direction[0])
        elif com == 'T':
            oi.append(Instruction('taper',(ival,),{'width': width}))
        elif com == '@':
            oi.append(Instruction('turn',(ival,),{'width': width}))
    return oi


import emerge as em
from emerge.plot import plot_sp
import numpy as np

""" COUPLED LINE FILTER DEMO

In this demo we construct a coupled-line bandpass filter using the
PCB Layouter interface in EMerge.
"""
# --- Unit definitions -----------------------------------------------------
mm = 0.001               # meter per millimeter
mil = 0.0254 * mm        # meter per mil (thousandth of an inch)

# --- Geometric parameters ------------------------------------------------
W = 20                   # reference trace width (mil)
D = 20                   # trace separation (mil)
w0 = 37                 # input/output line width (mil)
l0 = 100                # input/output line length (mil)
l1 = 314.22             # coupled section total length (mil)
l2 = 301.658            # inner coupled section length (mil)
l3 = 300.589            # outer coupled section length (mil)

# widths of each coupled segment (mil)
w1, w2, w3, w4, w5, w6 = 18.8, 43.484, 44.331, 44.331, 43.484, 18.8
# gaps between segments (mil)
g1, g2, g3, g4, g5, g6 = 9.63, 24.84, 41.499, 41.499, 24.84, 9.63

# substrate thickness (mil) and dielectric constant
th = 20
# dielectric constant of substrate (relative)
e = 0

# overall board dimensions/clearance (for bounding box)
Wtot = 2 * l1 + 5 * l2 + 7 * e + 2 * l0
WP = 200                # wide padding region (mil)
Dtot = 750              # total clearance (mil)
extra = 100             # extra margin (mil)

# --- Simulation setup ----------------------------------------------------
model = em.Simulation('Demo3', loglevel='DEBUG')

# --- Material and layouter -----------------------------------------------
mat = em.Material(er=3.55, color="#488343", opacity=0.1)
# Create PCB layouter with given substrate thickness and units
pcb = em.geo.PCB(th, unit=mil)

# --- Route coupled-line trace --------------------------------------------
# start at (0,140) with width w0
# label input port
pcb.new(0, 140, w0, (1, 0)) \
    .store('p1') \
    .straight(l0) \
    .turn(0) \
    .straight(l1 * 0.8) \
    .straight(l1, w1, dy=abs(w1 - w0) / 2) \
    .jump(gap=g1, side='left', reverse=l1 - e) \
    .straight(l1, w1) \
    .straight(l2, w2, dy=abs(w2 - w1) / 2) \
    .jump(gap=g2, side='left', reverse=l2 - e) \
    .straight(l2, w2) \
    .straight(l3, w3) \
    .jump(gap=g3, side='left', reverse=l2 - e) \
    .straight(l2, w3) \
    .straight(l3, w4) \
    .jump(gap=g4, side='left', reverse=l2 - e) \
    .straight(l2, w4) \
    .straight(l2, w5) \
    .jump(gap=g5, side='left', reverse=l2 - e) \
    .straight(l2, w5) \
    .straight(l1, w6, dy=abs(w2 - w1) / 2) \
    .jump(gap=g6, side='left', reverse=l1 - e) \
    .straight(l1, w6) \
    .turn(0) \
    .straight(l1 * 0.8, w0) \
    .straight(l0, w0, dy=abs(w1 - w0) / 2) \
    .store('p2')                        # label output port

# Compile the routed paths into a single GeoSurface
stripline = pcb.compile_paths(merge=True)

# --- Define simulation bounding box --------------------------------------
pcb.determine_bounds(topmargin=150, bottommargin=150)

# --- Generate dielectric and air blocks ----------------------------------
diel = pcb.generate_pcb()                     # substrate dielectric block
air = pcb.generate_air(4 * th)               # surrounding air box
# assign our custom material to the dielectric block
diel.material = mat

# --- Define ports for simulation ----------------------------------------
p1 = pcb.modal_port(pcb.load('p1'), width_multiplier=5, height=4 * th)
p2 = pcb.modal_port(pcb.load('p2'), width_multiplier=5, height=4 * th)

# --- Solver settings -----------------------------------------------------
model.mw.set_resolution(0.2)            # mesh density: fraction of wavelength
model.mw.set_frequency_range(5.2e9, 6.2e9, 31)  # 5.2–6.2 GHz, 31 points

# --- Assemble geometry into simulation -----------------------------------
model.commit_geometry(stripline, diel, p1, p2, air)

# --- Mesh refinement -----------------------------------------------------
model.mesher.set_boundary_size(stripline, 0.5 * mm)

# --- Mesh generation and view --------------------------------------------
model.generate_mesh()                    # build mesh
model.view(use_gmsh=True)                # visualize with Gmsh viewer

# --- Boundary conditions ------------------------------------------------
port1 = model.mw.bc.ModalPort(p1, 1, TEM=True)
port2 = model.mw.bc.ModalPort(p2, 2, TEM=True)
pec = model.mw.bc.PEC(stripline)

# --- Display setup -------------------------------------------------------
d = model.display
# show dielectric block translucent green and stripline in red
d.add_object(diel, color='green', opacity=0.5)
d.add_object(stripline)
d.show()

# --- Run frequency-domain solver ----------------------------------------
data = model.mw.run_sweep(parallel=True, njobs=4, frequency_groups=8)

# --- Extract and plot S-parameters ---------------------------------------
f = data.scalar.grid.freq                  # frequency axis
S11 = data.scalar.grid.S(1, 1)             # return loss
S21 = data.scalar.grid.S(2, 1)             # insertion loss
plot_sp(f / 1e9, [S11, S21], labels=['S11', 'S21'])

# --- Vector fitting and supersampled plot -------------------------------
f_fit = np.linspace(5.2e9, 6.2e9, 1001)
S11_fit = data.scalar.grid.model_S(1, 1, f_fit)
S21_fit = data.scalar.grid.model_S(2, 1, f_fit)
plot_sp(f_fit / 1e9, [S11_fit, S21_fit], labels=['S11', 'S21'])


# EMerge is an open source Python based FEM EM simulation module.
# Copyright (C) 2025  Robert Fennis.

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, see
# <https://www.gnu.org/licenses/>.

from .pcb import PCB
from .pmlbox import pmlbox
from .horn import Horn
from .shapes import Cylinder, CoaxCylinder, Box, XYPlate, HalfSphere, Sphere, Plate, OldBox, Alignment, Cone
from .operations import subtract, add, embed, remove, rotate, mirror, change_coordinate_system, translate, intersect
from .polybased import XYPolygon, GeoPrism, Disc, Curve
from .step import STEPItems


import emerge as em
import numpy as np
from emerge.plot import smith, plot_sp

""" STEPPED IMPEDANCE FILTER

In this demo we will look at how we can construct a stepped impedance filter using the
PCB Layouter interface in EMerge.

"""
# First we will define some constants/variables/parameters for our simulation.
mm = 0.001
mil = 0.0254*mm

L0, L1, L2, L3 = 400, 660, 660, 660 # The lengths of the sections in mil's
W0, W1, W2, W3 = 50, 128, 8, 224 # The widths of the sections in mil's

th = 62 # The PCB Thickness
er = 2.2 # The Dielectric constant

Hair = 60

## Material definition

# We can define the material using the Material class. Just supply the dielectric properties and you are done!
pcbmat = em.Material(er=2.2, tand=0.00, color="#217627", opacity=0.2)

# We start by creating our simulation object.

m = em.Simulation('Demo1_SIF', loglevel='INFO')

# To accomodate PCB routing we make use of the PCBLayouter class. To use it we need to 
# supply it with a thickness, the desired air-box height, the units at which we supply
# the dimensions and the PCB material.

layouter = em.geo.PCB(th, unit=mil, material=pcbmat, layers=3)

# We will route our PCB using the "method chaining" syntax. First we call the .new() method
# to start a new trace. This will returna StripPath object on which we may call methods that
# sequentially constructs our stripline trace. In this case, it is sipmly a sequence of straight
# sections.

layouter.new(0,0,W0, (1,0), z=layouter.z(2)).store('p1').straight(L0, W0).straight(L1,W1).straight(L2,W2).straight(L3,W3)\
    .straight(L2,W2).straight(L1,W1).straight(L0,W0).store('p2')

# Next we generate a wave port surface to use for our simulation. A wave port can be automatically
# generated for a given stripoline section. To easily reference it we use the .ref() method to 
# recall the sections we created earlier.
p1 = layouter.modal_port(layouter.load('p1'), height=0)
p2 = layouter.modal_port(layouter.load('p2'), height=0)

# Finally we compile the stirpline into a polygon. The compile_paths function will return
# GeoSurface objects that form the polygon. Additionally, we may turn on the Merge feature
# which will then return a single GeoSurface type object that we can use later.
polies = layouter.compile_paths(True)

# We can manually define blocks for the dielectric or air or let the PCBLayouter do it for us.
# First we must determine the bounds of our PCB. This function by default will make a PCB
# just large enough to contain all the coordinates in it (in the XY plane). By adding extra
# margins we can make sure to add sufficient space next to the trace. Just make sure that there
# is no margin where the wave ports need to go.
layouter.determine_bounds(leftmargin=0, topmargin=200, rightmargin=0, bottommargin=200)

# We can now generate the PCB and air box. The material assignment is automatic!

pcb = layouter.generate_pcb(True, merge=True)

# We now pass all the geometries we have created to the .commit_geometry() method.
m.commit_geometry()

# We set our desired resolution (fraction of the wavelength)
m.mw.set_resolution(0.15)

# And we define our frequency range
m.mw.set_frequency_range(0.2e9, 8e9, 41)

# EMerge also has a convenient interface to improve surface meshing quality. 
# With the set_boundary_size(method) we can define a meshing resolution for the edges of boundaries.
# This is adviced for small stripline structures.
# The growth_rate setting allows us to change how fast the mesh size will recover to the original size.
m.mesher.set_boundary_size(polies, 2*mm, growth_rate=1.2)
m.mesher.set_face_size(p1, 2*mm)
m.mesher.set_face_size(p2, 2*mm)

# Finally we generate our mesh and view it
m.generate_mesh()

m.view()

# We can now define the modal ports for the in and outputs and set the conductor to PEC.
port1 = m.mw.bc.ModalPort(p1, 1, TEM=True)
port2 = m.mw.bc.ModalPort(p2, 2, TEM=True)
pec = m.mw.bc.PEC(polies)

## OPTIONAL
# If we want to view the port mode we have to first know it. The modes are computed using a modal analysis.
# Since the latest version of EMerge. Modal analysis is executed automatically. Here we compute them manually
# as the mode is only calculated during the frequency domain solution.
if False:
    # Make sure to set the TEM property to True so that
    # EMerge knows to handle the port mode as a TEM boundary. This also includes the automatic
    # determination of a voltage integration line used for computing the port impedance.
    m.mw.modal_analysis(port1, 1, TEM=True)
    m.mw.modal_analysis(port2, 1, TEM=True)

    # Finally we import the display class to view the resultant modes
    m.display.add_object(pcb, opacity=0.1)
    m.display.add_object(polies)
    m.display.add_portmode(port1, 21)
    m.display.add_portmode(port2, 21)
    m.display.show()

# Finally we execute the frequency domain sweep and compute the Scattering Parameters.
sol = m.mw.run_sweep(parallel=True, njobs=4, frequency_groups=8)

# Our "sol" variable is of type MWData (Microwave Data). This contains a set of scalar data 
# like S-parameters and field data like the E/H field. The scalar data is in sol.scalar and the 
# field data in sol.field. Our data is currently a large list of entries in the dataset simply in order
# at which it is computed. We can structure our data in cases where we do a single frequency sweep
# or with a structured parameter sweep. In this case we can use the .grid property which will attempt
# to construct an N-dimensional grid from our results. 

gritted_data = sol.scalar.grid

# The gritted_data is of type MWGlobalNdim which means its an N-dimensional set of data.
# From this we can simply take all that we need.

f = gritted_data.freq
S11 = gritted_data.S(1,1)
S21 = gritted_data.S(2,1)

# This extracts the actual simulation data.
plot_sp(f/1e9, [S11, S21], labels=['S11','S21'], dblim=[-40,6], logx=True)

# We can also supersample our data by constructing a model using the Vector Fitting algorithm

f = np.linspace(0.2e9, 8e9, 2001)
S11 = gritted_data.model_S(1,1,f)
S21 = gritted_data.model_S(2,1,f)

smith(S11, labels='S11', f=f)

plot_sp(f/1e9, [S11, S21], labels=['S11','S21'], dblim=[-40,6], logx=True)

m.display.add_object(pcb, opacity=0.1)
m.display.add_object(polies, opacity=0.5)
m.display.add_surf(*sol.field[0].cutplane(1*mm, z=-0.75*th*mil).scalar('Ez','real'))
m.display.show()


# EMerge is an open source Python based FEM EM simulation module.
# Copyright (C) 2025  Robert Fennis.

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, see
# <https://www.gnu.org/licenses/>.

import numpy as np
from ....elements.nedleg2 import NedelecLegrange2
from scipy.sparse import csr_matrix
from ....mth.optimized import local_mapping, matinv, compute_distances, gaus_quad_tri
from numba import c16, types, f8, i8, njit, prange



############################################################
#                      FIELD MAPPING                      #
############################################################

@njit(i8[:,:](i8, i8[:,:], i8[:,:], i8[:,:]), cache=True, nogil=True)
def local_tri_to_edgeid(itri: int, tris, edges, tri_to_edge) -> np.ndarray:
    global_edge_map = edges[:, tri_to_edge[:,itri]]
    return local_mapping(tris[:, itri], global_edge_map)



############################################################
#                     PYTHON INTERFACE                     #
############################################################

def generelized_eigenvalue_matrix(field: NedelecLegrange2,
                           er: np.ndarray, 
                           ur: np.ndarray,
                           basis: np.ndarray,
                           k0: float,) -> tuple[csr_matrix, csr_matrix]:
    
    tris = field.mesh.tris
    edges = field.mesh.edges
    nodes = field.mesh.nodes
    
    nT = tris.shape[1]
    tri_to_field = field.tri_to_field

    nodes = field.local_nodes
    
    dataE, dataB, rows, cols = _matrix_builder(nodes, tris, edges, tri_to_field, ur, er, k0)
    
    nfield = field.n_field

    E = csr_matrix((dataE, (rows, cols)), shape=(nfield, nfield))
    B = csr_matrix((dataB, (rows, cols)), shape=(nfield, nfield))

    return E, B


############################################################
#                   MATRIX MULTIPLICATION                  #
############################################################

@njit(c16[:,:](c16[:,:], c16[:,:]), cache=True, nogil=True)
def matmul(a, b):
    out = np.empty((2,b.shape[1]), dtype=np.complex128)
    out[0,:] = a[0,0]*b[0,:] + a[0,1]*b[1,:]
    out[1,:] = a[1,0]*b[0,:] + a[1,1]*b[1,:]
    return out


############################################################
#              GAUSS QUADRATURE IMPLEMENTATION             #
############################################################

@njit(c16(c16[:], c16[:], types.Array(types.float64, 1, 'A', readonly=True)), cache=True, nogil=True)
def _gqi(v1, v2, W):
    return np.sum(v1*v2*W)

@njit(c16(c16[:,:], c16[:,:], types.Array(types.float64, 1, 'A', readonly=True)), cache=True, nogil=True)
def _gqi2(v1, v2, W):
    return np.sum(W*np.sum(v1*v2,axis=0))

@njit(c16[:,:](f8[:,:], f8[:,:]), cache=True, nogil=True)
def _ne1(coeff, coords):
    a1, b1, c1 = coeff[:,0]
    a2, b2, c2 = coeff[:,1]
    xs = coords[0,:]
    ys = coords[1,:]
    out=np.empty((2,xs.shape[0]), dtype=np.complex128)
    out[0,:] = (b1*(a2 + b2*xs + c2*ys) - b2*(a1 + b1*xs + c1*ys))*(a1 + b1*xs + c1*ys)
    out[1,:] = (c1*(a2 + b2*xs + c2*ys) - c2*(a1 + b1*xs + c1*ys))*(a1 + b1*xs + c1*ys)
    return out

@njit(c16[:,:](f8[:,:], f8[:,:]), cache=True, nogil=True)
def _ne2(coeff, coords):
    a1, b1, c1 = coeff[:,0]
    a2, b2, c2 = coeff[:,1]
    xs = coords[0,:]
    ys = coords[1,:]
    out=np.empty((2,xs.shape[0]), dtype=np.complex128)
    out[0,:] = (b1*(a2 + b2*xs + c2*ys) - b2*(a1 + b1*xs + c1*ys))*(a2 + b2*xs + c2*ys)
    out[1,:] = (c1*(a2 + b2*xs + c2*ys) - c2*(a1 + b1*xs + c1*ys))*(a2 + b2*xs + c2*ys)
    return out

@njit(c16[:,:](f8[:,:], f8[:,:]), cache=True, nogil=True)
def _nf1(coeff, coords):
    a1, b1, c1 = coeff[:,0]
    a2, b2, c2 = coeff[:,1]
    a3, b3, c3 = coeff[:,2]
    xs = coords[0,:]
    ys = coords[1,:]
    out=np.empty((2,xs.shape[0]), dtype=np.complex128)
    out[0,:] = -(b1*(a3 + b3*xs + c3*ys) - b3*(a1 + b1*xs + c1*ys))*(a2 + b2*xs + c2*ys)
    out[1,:] = -(c1*(a3 + b3*xs + c3*ys) - c3*(a1 + b1*xs + c1*ys))*(a2 + b2*xs + c2*ys)
    return out

@njit(c16[:,:](f8[:,:], f8[:,:]), cache=True, nogil=True)
def _nf2(coeff, coords):
    a1, b1, c1 = coeff[:,0]
    a2, b2, c2 = coeff[:,1]
    a3, b3, c3 = coeff[:,2]
    xs = coords[0,:]
    ys = coords[1,:]
    out=np.empty((2,xs.shape[0]), dtype=np.complex128)
    out[0,:] = (b1*(a2 + b2*xs + c2*ys) - b2*(a1 + b1*xs + c1*ys))*(a3 + b3*xs + c3*ys)
    out[1,:] = (c1*(a2 + b2*xs + c2*ys) - c2*(a1 + b1*xs + c1*ys))*(a3 + b3*xs + c3*ys)
    return out

@njit(c16[:](f8[:], f8[:,:]), cache=True, nogil=True)
def _lv(coeff, coords):
    a1, b1, c1 = coeff
    xs = coords[0,:]
    ys = coords[1,:]
    return -a1 - b1*xs - c1*ys + 2*(a1 + b1*xs + c1*ys)**2 + 0*1j

@njit(c16[:](f8[:,:], f8[:,:]), cache=True, nogil=True)
def _le(coeff, coords):
    a1, b1, c1 = coeff[:,0]
    a2, b2, c2 = coeff[:,1]
    xs = coords[0,:]
    ys = coords[1,:]
    return 4*(a1 + b1*xs + c1*ys)*(a2 + b2*xs + c2*ys)+ 0*1j

@njit(c16[:,:](f8[:], f8[:,:]), cache=True, nogil=True)
def _lv_grad(coeff, coords):
    a1, b1, c1 = coeff
    xs = coords[0,:]
    ys = coords[1,:]
    out=np.empty((2,xs.shape[0]), dtype=np.complex128)
    out[0,:] = b1*(4*a1 + 4*b1*xs + 4*c1*ys - 1)
    out[1,:] = c1*(4*a1 + 4*b1*xs + 4*c1*ys - 1)
    return out

@njit(c16[:,:](f8[:,:], f8[:,:]), cache=True, nogil=True)
def _le_grad(coeff, coords):
    a1, b1, c1 = coeff[:,0]
    a2, b2, c2 = coeff[:,1]
    xs = coords[0,:]
    ys = coords[1,:]
    out=np.empty((2,xs.shape[0]), dtype=np.complex128)
    out[0,:] = 4*b1*(a2 + b2*xs + c2*ys) + 4*b2*(a1 + b1*xs + c1*ys)
    out[1,:] = 4*c1*(a2 + b2*xs + c2*ys) + 4*c2*(a1 + b1*xs + c1*ys)
    return out

@njit(c16[:](f8[:,:], f8[:,:]), cache=True, nogil=True)
def _ne1_curl(coeff, coords):
    a1, b1, c1 = coeff[:,0]
    a2, b2, c2 = coeff[:,1]
    xs = coords[0,:]
    ys = coords[1,:]
    return -3*a1*b1*c2 + 3*a1*b2*c1 - 3*b1**2*c2*xs + 3*b1*b2*c1*xs - 3*b1*c1*c2*ys + 3*b2*c1**2*ys + 0j


@njit(c16[:](f8[:,:], f8[:,:]), cache=True, nogil=True)
def _ne2_curl(coeff, coords):
    a1, b1, c1 = coeff[:,0]
    a2, b2, c2 = coeff[:,1]
    xs = coords[0,:]
    ys = coords[1,:]
    return -3*a2*b1*c2 + 3*a2*b2*c1 - 3*b1*b2*c2*xs - 3*b1*c2**2*ys + 3*b2**2*c1*xs + 3*b2*c1*c2*ys+ 0j

@njit(c16[:](f8[:,:], f8[:,:]), cache=True, nogil=True)
def _nf1_curl(coeff, coords):
    a1, b1, c1 = coeff[:,0]
    a2, b2, c2 = coeff[:,1]
    a3, b3, c3 = coeff[:,2]
    xs = coords[0,:]
    ys = coords[1,:]
    return -b2*(c1*(a3 + b3*xs + c3*ys) - c3*(a1 + b1*xs + c1*ys)) + c2*(b1*(a3 + b3*xs + c3*ys) - b3*(a1 + b1*xs + c1*ys)) + 2*(b1*c3 - b3*c1)*(a2 + b2*xs + c2*ys) + 0*1j
           
@njit(c16[:](f8[:,:], f8[:,:]), cache=True, nogil=True)
def _nf2_curl(coeff, coords):
    a1, b1, c1 = coeff[:,0]
    a2, b2, c2 = coeff[:,1]
    a3, b3, c3 = coeff[:,2]
    xs = coords[0,:]
    ys = coords[1,:]
    return b3*(c1*(a2 + b2*xs + c2*ys) - c2*(a1 + b1*xs + c1*ys)) - c3*(b1*(a2 + b2*xs + c2*ys) - b2*(a1 + b1*xs + c1*ys)) - 2*(b1*c2 - b2*c1)*(a3 + b3*xs + c3*ys) + 0*1j
           

############################################################
#     TRIANGLE BARYCENTRIC COORDINATE LIN. COEFFICIENTS    #
############################################################


@njit(types.Tuple((f8[:], f8[:], f8[:], f8))(f8[:], f8[:]), cache = True, nogil=True)
def tri_coefficients(vxs, vys):

    x1, x2, x3 = vxs
    y1, y2, y3 = vys

    a1 = x2*y3-y2*x3
    a2 = x3*y1-y3*x1
    a3 = x1*y2-y1*x2
    b1 = y2-y3
    b2 = y3-y1
    b3 = y1-y2
    c1 = x3-x2
    c2 = x1-x3
    c3 = x2-x1

    #A = 0.5*(b1*c2 - b2*c1)
    sA = 0.5*(((x1-x3)*(y2-y1) - (x1-x2)*(y3-y1)))
    sign = np.sign(sA)
    A = np.abs(sA)
    As = np.array([a1, a2, a3])*sign
    Bs = np.array([b1, b2, b3])*sign
    Cs = np.array([c1, c2, c3])*sign
    return As, Bs, Cs, A


############################################################
#                    CONSTANT DEFINITION                   #
############################################################


DPTS = np.array([[0.22338159, 0.22338159, 0.22338159, 0.10995174, 0.10995174, 0.10995174],
                 [0.10810302, 0.44594849, 0.44594849, 0.81684757, 0.09157621, 0.09157621],
                 [0.44594849, 0.44594849, 0.10810302, 0.09157621, 0.09157621, 0.81684757],
                 [0.44594849, 0.10810302, 0.44594849, 0.09157621, 0.81684757, 0.09157621]], dtype=np.float64)


############################################################
#                 NUMBA OPTIMIZED ASSEMBLER                #
############################################################


@njit(types.Tuple((c16[:,:], c16[:,:]))(f8[:,:], i8[:,:], c16[:,:], c16[:,:], f8), cache=True, nogil=True)
def generalized_matrix_GQ(tri_vertices, local_edge_map, Ms, Mm, k0):
    '''Nedelec-2 Triangle stiffness and mass submatrix'''
    Att = np.zeros((8,8), dtype=np.complex128)
    Btt = np.zeros((8,8), dtype=np.complex128)

    Dtt = np.zeros((8,8), dtype=np.complex128)
    Dzt = np.zeros((6,8), dtype=np.complex128)

    Dzz1 = np.zeros((6,6), dtype=np.complex128)
    Dzz2 = np.zeros((6,6), dtype=np.complex128)
    
    Ls = np.ones((14,14), dtype=np.float64)

    WEIGHTS = DPTS[0,:]
    DPTS1 = DPTS[1,:]
    DPTS2 = DPTS[2,:]
    DPTS3 = DPTS[3,:]

    txs = tri_vertices[0,:]
    tys = tri_vertices[1,:]

    Ds = compute_distances(txs, tys, 0*txs)

    xs = txs[0]*DPTS1 + txs[1]*DPTS2 + txs[2]*DPTS3
    ys = tys[0]*DPTS1 + tys[1]*DPTS2 + tys[2]*DPTS3
    
    cs = np.empty((2,xs.shape[0]), dtype=np.float64)
    cs[0,:] = xs
    cs[1,:] = ys

    aas, bbs, ccs, Area = tri_coefficients(txs, tys)

    coeff = np.empty((3,3), dtype=np.float64)
    coeff[0,:] = aas/(2*Area)
    coeff[1,:] = bbs/(2*Area)
    coeff[2,:] = ccs/(2*Area)

    Msz = Ms[2,2]
    Mmz = Mm[2,2]
    Ms = Ms[:2,:2]
    Mm = Mm[:2,:2]

    Ls[3,:] *= Ds[0,2]
    Ls[7,:] *= Ds[0,1]
    Ls[:,3] *= Ds[0,2]
    Ls[:,7] *= Ds[0,1]
    
    for iv1 in range(3):
        ie1 = local_edge_map[:, iv1]
        
        Le = Ds[ie1[0], ie1[1]]
        Ls[iv1,:] *= Le
        Ls[:,iv1] *= Le
        Ls[iv1+4,:] *= Le
        Ls[:,iv1+4] *= Le
        F1 = _ne1_curl(coeff[:,ie1], cs)
        F2 = _ne2_curl(coeff[:,ie1], cs)
        F3 = _ne1(coeff[:,ie1], cs)
        F4 = _ne2(coeff[:,ie1], cs)
        F5 = _lv_grad(coeff[:,iv1],cs)
        F6 = _le_grad(coeff[:,ie1],cs)
        
        for iv2 in range(3):
            ei2 = local_edge_map[:, iv2]
            
            H1 = matmul(Ms,_ne1(coeff[:,ei2],cs))
            H2 = matmul(Ms,_ne2(coeff[:,ei2],cs))
            
            Att[iv1,iv2]     = _gqi(F1, Msz * _ne1_curl(coeff[:,ei2],cs), WEIGHTS)
            Att[iv1+4,iv2]   = _gqi(F2, Msz * _ne1_curl(coeff[:,ei2],cs), WEIGHTS)
            Att[iv1,iv2+4]   = _gqi(F1, Msz * _ne2_curl(coeff[:,ei2],cs), WEIGHTS)
            Att[iv1+4,iv2+4] = _gqi(F2, Msz * _ne2_curl(coeff[:,ei2],cs), WEIGHTS)

            Btt[iv1,iv2]     = _gqi2(F3, matmul(Mm,_ne1(coeff[:,ei2],cs)), WEIGHTS)
            Btt[iv1+4,iv2]   = _gqi2(F4, matmul(Mm,_ne1(coeff[:,ei2],cs)), WEIGHTS)
            Btt[iv1,iv2+4]   = _gqi2(F3, matmul(Mm,_ne2(coeff[:,ei2],cs)), WEIGHTS)
            Btt[iv1+4,iv2+4] = _gqi2(F4, matmul(Mm,_ne2(coeff[:,ei2],cs)), WEIGHTS)

            Dtt[iv1,iv2]     = _gqi2(F3, H1, WEIGHTS)
            Dtt[iv1+4,iv2]   = _gqi2(F4, H1, WEIGHTS)
            Dtt[iv1,iv2+4]   = _gqi2(F3, H2, WEIGHTS)
            Dtt[iv1+4,iv2+4] = _gqi2(F4, H2, WEIGHTS)

            Dzt[iv1, iv2]     = _gqi2(F5, H1, WEIGHTS)
            Dzt[iv1+3, iv2]   = _gqi2(F6, H1, WEIGHTS)
            Dzt[iv1, iv2+4]   = _gqi2(F5, H2, WEIGHTS)
            Dzt[iv1+3, iv2+4] = _gqi2(F6, H2, WEIGHTS)

            Dzz1[iv1, iv2]     = _gqi2(_lv_grad(coeff[:,iv1], cs), matmul(Ms,_lv_grad(coeff[:,iv2],cs)), WEIGHTS)
            Dzz1[iv1, iv2+3]   = _gqi2(_lv_grad(coeff[:,iv1], cs), matmul(Ms,_le_grad(coeff[:,ei2],cs)), WEIGHTS)
            Dzz1[iv1+3, iv2]   = _gqi2(_le_grad(coeff[:,ie1], cs), matmul(Ms,_lv_grad(coeff[:,iv2],cs)), WEIGHTS)
            Dzz1[iv1+3, iv2+3] = _gqi2(_le_grad(coeff[:,ie1], cs), matmul(Ms,_le_grad(coeff[:,ei2],cs)), WEIGHTS)

            Dzz2[iv1, iv2]     = _gqi(_lv(coeff[:,iv1], cs), Mmz * _lv(coeff[:,iv2],cs), WEIGHTS)
            Dzz2[iv1, iv2+3]   = _gqi(_lv(coeff[:,iv1], cs), Mmz * _le(coeff[:,ei2],cs), WEIGHTS)
            Dzz2[iv1+3, iv2]   = _gqi(_le(coeff[:,ie1], cs), Mmz * _lv(coeff[:,iv2],cs), WEIGHTS)
            Dzz2[iv1+3, iv2+3] = _gqi(_le(coeff[:,ie1], cs), Mmz * _le(coeff[:,ei2],cs), WEIGHTS)


        G1 = matmul(Mm,_nf1(coeff,cs))
        G2 = matmul(Mm,_nf2(coeff,cs))
        G3 = matmul(Ms,_nf1(coeff,cs))
        G4 = matmul(Ms,_nf2(coeff,cs))
        
        Att[iv1,3]   = _gqi(F1, Msz * _nf1_curl(coeff,cs), WEIGHTS)
        Att[iv1+4,3] = _gqi(_ne2_curl(coeff[:,ie1], cs), Msz * _nf1_curl(coeff,cs), WEIGHTS)
        Att[iv1,7]   = _gqi(F1, Msz * _nf2_curl(coeff,cs), WEIGHTS)
        Att[iv1+4,7] = _gqi(_ne2_curl(coeff[:,ie1], cs), Msz * _nf2_curl(coeff,cs), WEIGHTS)
        
        Att[3, iv1]   = Att[iv1,3]
        Att[7, iv1]   = Att[iv1,7]
        Att[3, iv1+4] = Att[iv1+4,3]
        Att[7, iv1+4] = Att[iv1+4,7]

        Btt[iv1,3]   = _gqi2(F3, G1, WEIGHTS)
        Btt[iv1+4,3] = _gqi2(F4, G1, WEIGHTS)
        Btt[iv1,7]   = _gqi2(F3, G2, WEIGHTS)
        Btt[iv1+4,7] = _gqi2(F4, G2, WEIGHTS)

        Btt[3, iv1]   = Btt[iv1,3]
        Btt[7, iv1]   = Btt[iv1,7]
        Btt[3, iv1+4] = Btt[iv1+4,3]
        Btt[7, iv1+4] = Btt[iv1+4,7]

        Dtt[iv1,3]   = _gqi2(F3, G3, WEIGHTS)
        Dtt[iv1+4,3] = _gqi2(F4, G3, WEIGHTS)
        Dtt[iv1,7]   = _gqi2(F3, G4, WEIGHTS)
        Dtt[iv1+4,7] = _gqi2(F4, G4, WEIGHTS)

        Dtt[3, iv1]   = Dtt[iv1,3]
        Dtt[7, iv1]   = Dtt[iv1,7]
        Dtt[3, iv1+4] = Dtt[iv1+4,3]
        Dtt[7, iv1+4] = Dtt[iv1+4,7]

        Dzt[iv1, 3]   = _gqi2(F5, G3, WEIGHTS)
        Dzt[iv1, 7]   = _gqi2(F5, G4, WEIGHTS)
        Dzt[iv1+3, 3] = _gqi2(F6, G3, WEIGHTS)
        Dzt[iv1+3, 7] = _gqi2(F6, G4, WEIGHTS)

    Att[3,3] = _gqi(_nf1_curl(coeff, cs), Msz * _nf1_curl(coeff,cs), WEIGHTS)
    Att[7,3] = _gqi(_nf2_curl(coeff, cs), Msz * _nf1_curl(coeff,cs), WEIGHTS)
    Att[3,7] = _gqi(_nf1_curl(coeff, cs), Msz * _nf2_curl(coeff,cs), WEIGHTS)
    Att[7,7] = _gqi(_nf2_curl(coeff, cs), Msz * _nf2_curl(coeff,cs), WEIGHTS)

    Btt[3,3] = _gqi2(_nf1(coeff, cs), G1, WEIGHTS)
    Btt[7,3] = _gqi2(_nf2(coeff, cs), G1, WEIGHTS)
    Btt[3,7] = _gqi2(_nf1(coeff, cs), G2, WEIGHTS)
    Btt[7,7] = _gqi2(_nf2(coeff, cs), G2, WEIGHTS)

    A = np.zeros((14, 14), dtype = np.complex128)
    B = np.zeros((14, 14), dtype = np.complex128)
    
    A[:8,:8] = (Att - k0**2 * Btt)

    B[:8,:8] = Dtt
    B[8:,:8] = Dzt
    B[:8,8:] = Dzt.T
    B[8:,8:] = Dzz1 - k0**2 * Dzz2

    #Ls = np.ones((14,14), dtype=np.float64)
    
    B = Ls*B*np.abs(Area)
    A = Ls*A*np.abs(Area)
    return A, B


@njit(types.Tuple((c16[:], c16[:], i8[:], i8[:]))(f8[:,:],
                                                      i8[:,:], 
                                                      i8[:,:], 
                                                      i8[:,:],
                                                      c16[:,:,:], 
                                                      c16[:,:,:], 
                                                      f8), cache=True, nogil=True, parallel=True)
def _matrix_builder(nodes, tris, edges, tri_to_field, ur, er, k0):

    ntritot = tris.shape[1]
    nnz = ntritot*196

    rows = np.zeros(nnz, dtype=np.int64)
    cols = np.zeros(nnz, dtype=np.int64)
    dataE = np.zeros_like(rows, dtype=np.complex128)
    dataB = np.zeros_like(rows, dtype=np.complex128)

    tri_to_edge = tri_to_field[:3,:]
    
    for itri in prange(ntritot): # type: ignore
        p = itri*196
        urt = ur[:,:,itri]
        ert = er[:,:,itri]

        # Construct a local mapping to global triangle orientations
        local_tri_map = local_tri_to_edgeid(itri, tris, edges, tri_to_edge)

        # Construct the local edge map
        tri_nodes = nodes[:, tris[:,itri]]
        Esub, Bsub = generalized_matrix_GQ(tri_nodes,local_tri_map, matinv(urt), ert, k0)
        
        indices = tri_to_field[:, itri]
        for ii in range(14):
            rows[p+14*ii:p+14*(ii+1)] = indices[ii]
            cols[p+ii:p+ii+196:14] = indices[ii]

        dataE[p:p+196] = Esub.ravel()
        dataB[p:p+196] = Bsub.ravel()
    return dataE, dataB, rows, cols


# EMerge is an open source Python based FEM EM simulation module.
# Copyright (C) 2025  Robert Fennis.

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, see
# <https://www.gnu.org/licenses/>.

from .cs import Axis, _parse_axis, GCS, _parse_vector
from .selection import SELECTOR_OBJ, Selection
from .geo import GeoPrism, XYPolygon, Alignment, XYPlate
from .bc import BoundaryCondition
from typing import Generator
from .bc import Periodic
import numpy as np


############################################################
#                         FUNCTIONS                        #
############################################################

def _rotnorm(v: np.ndarray) -> np.ndarray:
    """Rotate 3D vector field v 90° counterclockwise around z axis.

    v shape = (3, Ny, Nx)
    """
    ax = np.array([-v[1], v[0], v[2]])
    ax = ax/np.linalg.norm(ax)
    return ax

def _pair_selection(f1: Selection, 
                    f2: Selection, 
                    translation: tuple[float, float, float]):
    if len(f1.tags) == 1:
        return [f1,], [f2,]
    c1s = [np.array(c) for c in f1.centers]
    c2s = [np.array(c) for c in f2.centers]
    ds = np.array(translation)
    f1s = []
    f2s = []
    for t1, c1 in zip(f1.tags, c1s):
        for t2, c2 in zip(f2.tags, c2s):
            if np.linalg.norm((c1 + ds)-c2) < 1e-8:
                f1s.append(Selection([t1,]))
                f2s.append(Selection([t2,]))
    return f1s, f2s



############################################################
#                 BASE PERIODIC CELL CLASS                #
############################################################

# TODO: This must be moved to mw physics if possible
class PeriodicCell:

    def __init__(self, 
                 origins: list[tuple[float, float, float] | list[float] | np.ndarray | Axis],
                 vectors: list[tuple[float, float, float] | list[float] | np.ndarray | Axis]):

        self.origins: list[tuple[float, float, float]] = [_parse_vector(origin) for origin in origins] # type: ignore
        self.vectors: list[Axis] = [_parse_axis(vec) for vec in vectors]
        self.excluded_faces: Selection | None = None
        self._bcs: list[Periodic] = []
        self._ports: list[BoundaryCondition] = []

        self.__post_init__()

    def __post_init__(self):
        pass

    def volume(self, z1: float, z2: float) -> GeoPrism:
        """Genereates a volume with the cell geometry ranging from z1 tot z2

        Args:
            z1 (float): The start height
            z2 (float): The end height

        Returns:
            GeoPrism: The resultant prism
        """
        raise NotImplementedError('This method is not implemented for this subclass.')
    
    def cell_data(self) -> Generator[tuple[Selection, Selection, np.ndarray], None, None]:
        """An iterator that yields the two faces of the hex cell plus a cell periodicity vector

        Yields:
            Generator[np.ndarray, np.ndarray, np.ndarray]: The face and periodicity data
        """
        raise NotImplementedError('This method is not implemented for this subclass.')

    def generate_bcs(self) -> list[Periodic]:
        """Generate the priodic boundary conditions
        """
        bcs = []
        for f1, f2, a in self.cell_data():
            f1_new = f1
            f2_new = f2
            if self.excluded_faces is not None:
                f1_new = f1 - self.excluded_faces # type: ignore
                f2_new = f2 - self.excluded_faces # type: ignore
            bcs.append(Periodic(f1_new, f2_new, tuple(a)))
        self._bcs = bcs
        return bcs
    
    @property
    def bcs(self) -> list[BoundaryCondition]:
        """Returns a list of Periodic boundary conditions for the given PeriodicCell

        Args:
            exclude_faces (list[FaceSelection], optional): A possible list of faces to exclude from the bcs. Defaults to None.

        Returns:
            list[Periodic]: The list of Periodic boundary conditions
        """
        if not self._bcs:
            raise ValueError('Periodic Boundary conditions not generated')
        return self._bcs + self._ports # type: ignore
    
    def set_scanangle(self, theta: float, phi: float, degree: bool = True) -> None:
        """Sets the scanangle for the periodic condition. (0,0) is defined along the Z-axis

        Args:
            theta (float): The theta angle
            phi (float): The phi angle
            degree (bool): If the angle is in degrees. Defaults to True
        """
        if degree:
            theta = theta*np.pi/180
            phi = phi*np.pi/180

        
        ux = np.sin(theta)*np.cos(phi)
        uy = np.sin(theta)*np.sin(phi)
        uz = np.cos(theta)
        for bc in self._bcs:
            bc.ux = ux
            bc.uy = uy
            bc.uz = uz
        for port in self._ports:
            port.scan_theta = theta # type: ignore
            port.scan_phi = phi # type: ignore

    def port_face(self, z: float):
        raise NotImplementedError('')
    


############################################################
#                    RECTANGULAR TILING                   #
############################################################

class RectCell(PeriodicCell):
    """This class represents the unit cell environment of a regular rectangular tiling.

    Args:
        PeriodicCell (_type_): _description_
    """
    def __init__(self, 
                 width: float,
                 height: float,):
        """The RectCell class represents a regular rectangular tiling in the XY plane where
        the width is along the X-axis (centered at x=0) and the height along the Y-axis (centered at y=0)

        Args:
            width (float): The Cell width
            height (float): The Cell height
        """
        v1 = (width, 0, 0)
        o1 = (-width/2, 0, 0)
        v2 = (0, height, 0)
        o2 = (0, -height/2, 0)
        super().__init__([o1, o2], [v1, v2])
        self.width: float = width
        self.height: float = height
        self.fleft = (o1, v1)
        self.fbot = (o2, v2)
        self.ftop = ((0, height/2, 0), v2)
        self.fright = ((width/2, 0, 0), v1)

    def port_face(self, z: float):
        return XYPlate(self.width, self.height, position=(0,0,z), alignment=Alignment.CENTER)
        
    def cell_data(self):
        f1s = SELECTOR_OBJ.inplane(*self.fleft[0], *self.fleft[1])
        f2s = SELECTOR_OBJ.inplane(*self.fright[0], *self.fright[1])
        vec = (self.fright[0][0]-self.fleft[0][0], 
               self.fright[0][1]-self.fleft[0][1], 
               self.fright[0][2]-self.fleft[0][2])
        for f1, f2 in zip(*_pair_selection(f1s, f2s, vec)):
            yield f1, f2, vec

        f1s = SELECTOR_OBJ.inplane(*self.fbot[0], *self.fbot[1])
        f2s = SELECTOR_OBJ.inplane(*self.ftop[0], *self.ftop[1])
        vec = (self.ftop[0][0]-self.fbot[0][0], 
               self.ftop[0][1]-self.fbot[0][1], 
               self.ftop[0][2]-self.fbot[0][2])
        for f1, f2 in zip(*_pair_selection(f1s, f2s, vec)):
            yield f1, f2, vec

    def volume(self, 
               z1: float,
               z2: float) -> GeoPrism:
        xs = np.array([-self.width/2, self.width/2, self.width/2, -self.width/2])
        ys = np.array([-self.height/2, -self.height/2, self.height/2, self.height/2])
        poly = XYPolygon(xs, ys)
        length = z2-z1
        return poly.extrude(length, cs=GCS.displace(0,0,z1))
    


############################################################
#                     HEXAGONAL TILING                    #
############################################################

class HexCell(PeriodicCell):

    def __init__(self,
                 point1: tuple[float, float, float],
                 point2: tuple[float, float, float],
                 point3: tuple[float, float, float]):
        """
        Generates a Hexagonal periodic tiling by providing three coordinates.
        The layout of the tiling assumes a hexagon with a single vertex at the top and bottom,
        and one vertex on the bottom and right faces (⬢).

        Args:
            point1 (tuple[float, float, float]): left face top vertex
            point2 (tuple[float, float, float]): left face bottom vertex
            point3 (tuple[float, float, float]): bottom vertex
        """
        p1, p2, p3 = np.array(point1), np.array(point2), np.array(point3)
        p4 = -p1
        self.p1: np.ndarray = p1
        self.p2: np.ndarray = p2
        self.p3: np.ndarray = p3
        o1 = (p1+p2)/2
        o2 = (p2+p3)/2
        o3 = (p3+p4)/2
        self.o1 = o1
        self.o2 = o2
        self.o3 = o3
        n1 = _rotnorm(p2-p1)
        n2 = _rotnorm(p3-p2)
        n3 = _rotnorm(p4-p3)
        
        super().__init__([o1, o2, o3], [n1,n2,n3])

        self.f11 = (o1, n1)
        self.f21 = (o2, n2)
        self.f31 = (o3, n3)
        self.f12 = (-o1, n1)
        self.f22 = (-o2, n2)
        self.f32 = (-o3, n3)

    def port_face(self, z: float):
        xs, ys, zs = zip(self.p1, self.p2, self.p3)
        poly = XYPolygon(xs, ys).geo(GCS.displace(0,0,zs[0]))
        return poly
    
    def cell_data(self) -> Generator[tuple[Selection, Selection, np.ndarray], None, None]:
        nrm = np.linalg.norm

        o = self.o1[:-1]
        n = self.f11[1][:-1]
        w = nrm(self.p2-self.p1)/2
        f1s = SELECTOR_OBJ.inplane(*self.f11[0], *self.f11[1])\
            .exclude(lambda x, y, z: (nrm(np.array([x,y])-o)>w) or (abs((np.array([x,y])-o) @ n ) > 1e-6))
        f2s = SELECTOR_OBJ.inplane(*self.f12[0], *self.f12[1])\
            .exclude(lambda x, y, z: (nrm(np.array([x,y])+o)>w) or (abs((np.array([x,y])+o) @ n ) > 1e-6))
        vec = - (self.p1 + self.p2)

        for f1, f2 in zip(*_pair_selection(f1s, f2s, vec)): # type: ignore
            yield f1, f2, vec

        o = self.o2[:-1]
        n = self.f21[1][:-1]
        w = nrm(self.p3-self.p2)/2
        f1s = SELECTOR_OBJ.inplane(*self.f21[0], *self.f21[1])\
            .exclude(lambda x, y, z: (nrm(np.array([x,y])-o)>w) or (abs((np.array([x,y])-o) @ n ) > 1e-6))
        f2s = SELECTOR_OBJ.inplane(*self.f22[0], *self.f22[1])\
            .exclude(lambda x, y, z: (nrm(np.array([x,y])+o)>w) or (abs((np.array([x,y])+o) @ n ) > 1e-6))
        vec = - (self.p2 + self.p3)
        for f1, f2 in zip(*_pair_selection(f1s, f2s, vec)): # type: ignore
            yield f1, f2, vec
        
        o = self.o3[:-1]
        n = self.f31[1][:-1]
        w = nrm(-self.p1-self.p3)/2
        f1s = SELECTOR_OBJ.inplane(*self.f31[0], *self.f31[1])\
            .exclude(lambda x, y, z: (nrm(np.array([x,y])-o)>w) or (abs((np.array([x,y])-o) @ n ) > 1e-6))
        f2s = SELECTOR_OBJ.inplane(*self.f32[0], *self.f32[1])\
            .exclude(lambda x, y, z: (nrm(np.array([x,y])+o)>w) or (abs((np.array([x,y])+o) @ n ) > 1e-6))
        vec = - (self.p3 - self.p1)
        for f1, f2 in zip(*_pair_selection(f1s, f2s, vec)): # type: ignore
            yield f1, f2, vec

    def volume(self, 
               z1: float,
               z2: float) -> GeoPrism:
        xs, ys, zs = zip(self.p1, self.p2, self.p3)
        xs2 = np.array(xs) # type: ignore
        ys2 = np.array(ys) # type: ignore
        xs3 = np.concatenate([xs2, -xs2]) # type: ignore
        ys3 = np.concatenate([ys2, -ys2]) # type: ignore
        poly = XYPolygon(xs3, ys3)
        length = z2-z1
        return poly.extrude(length, cs=GCS.displace(0,0,z1))
    


# -----------------------------------------------------------------------------
# HELIX ANTENNA DEMO — Simulation setup, mesh, sweep, and far-field plots
#
# This script builds a helical radiator with a short feed, places it in an
# airbox, applies a lumped port and absorbing boundary, runs a frequency sweep,
# and visualizes S-parameters and radiation patterns (2D cuts + 3D).
# -----------------------------------------------------------------------------

import emerge as em
from emerge.plot import plot_sp, plot_ff, smith
import numpy as np

# --- Units & basic constants -------------------------------------------------
mm = 0.001                      # meters per millimeter

f0 = 3e9                        # center frequency (Hz)
c0 = 299792458                  # speed of light (m/s)
wl = c0/f0                      # free-space wavelength at f0 (m)
rad0 = wl/(2*3.1415)            # reference radius ~= λ/(2π)
radw = 1*mm                     # wire/pipe radius for the helix

L = 4*rad0                      # helix axial length

porth = 2*mm                    # vertical height of the feed extrusion

# --- Simulation object -------------------------------------------------------
model = em.Simulation('helix', loglevel="DEBUG")
dfeed = 5*mm                    # straight feed length before the helix starts

# --- Geometry: helix curve and metal pipe -----------------------------------
# Helix curve from (0,0,porth+dfeed/2) to (0,0,porth+dfeed/2+L)
# with initial radius rad0, 13 degree pitch, tapered end radius r_end
# The 'startfeed' parameter is radius like parameter to turn the start of the helix downward.
#
# This is illustrated in the diagrom below. Normally a helix would start at x1, y1.ArithmeticError
# If some start startfeed distance is provided, the feed will curve downwards adding an extra distance
# y1-y0 equal to startfeed/2. You can access the coordinate x0,0 usinging curve_obj.p0
#
#             ______
#           /       `-- (x1,y1)  --> helix
#         /                
#       /                  
#      |                   
#   (x0,y1)               

# We add the porth height + half of the startfeed distance to the total height to put the helix in the right spot.
h_curve = em.geo.Curve.helix_lh((0,0,porth+dfeed/2), (0,0,porth+dfeed/2+L), rad0, 13, r_end=0.8*rad0, startfeed=dfeed)

# Sweep a circular cross-section along the curve to make a metallic pipe
helix = h_curve.pipe(em.geo.XYPolygon.circle(radw, Nsections=6)).set_material(em.lib.MET_COPPER)

# Optional preview of current scene (geometry only at this point)
model.view()

# --- Meshing preferences on helix -------------------------------------------
helix.max_meshsize = 3*radw     # set a courser mesh size for the helix

# --- Feed geometry: short vertical extrusion at the base ---------------------
x0, y0, z0 = h_curve.p0         # helix start point x and y coordinate
feed_poly = em.geo.XYPolygon.circle(radw, Nsections=6)
feed = feed_poly.extrude(porth, em.GCS.displace(x0, y0, 0))  # vertical feed stub

# --- Background domain (airbox) & frequency sweep ---------------------------
# Airbox centered around the helix with some clearance in X/Y and height in Z
airbox = em.geo.Box(4*rad0, 4*rad0, 1.5*L, (-2*rad0, -2*rad0, 0)).background()

# Sweep across 2.8–3.4 GHz (11 points) to cover the operating band
model.mw.set_frequency_range(2.8e9, 3.4e9, 11)

# --- Mesh generation & preview ----------------------------------------------
model.generate_mesh()
model.view()

# --- Boundary selections for BCs & port -------------------------------------
abc_sel = airbox.outside('bottom')     # absorbing boundary applied at bottom face of airbox
port_sel = feed.outside('front','back')# port faces on the feed (two opposite faces)

# --- Boundary conditions -----------------------------------------------------
abc = model.mw.bc.AbsorbingBoundary(abc_sel)                     # open-space termination
port_sel = model.mw.bc.LumpedPort(port_sel, 1, feed_poly.length, porth, em.ZAX ,Z0=130)  # lumped port (Z-axis)

# --- Solve frequency sweep ---------------------------------------------------
data = model.mw.run_sweep()

# --- S-parameters: access and plots -----------------------------------------
glob = data.scalar.grid
plot_sp(glob.freq, glob.S(1,1))         # return loss vs frequency
smith(glob.S(1,1), f=glob.freq)         # Smith chart of S11

# --- Far-field 2D cuts (two principal planes) --------------------------------
# Cut 1: elevation sweep with phi defined by basis (0,0,1) and (0,1,0)
ff1 = data.field.find(freq=3.1e9).farfield_2d((0,0,1), (0,1,0), abc_sel, (-90, 90))
# Cut 2: elevation sweep with orthogonal basis (0,0,1) and (1,0,0)
ff2 = data.field.find(freq=3.1e9).farfield_2d((0,0,1), (1,0,0), abc_sel, (-90, 90))

# Normalize circular components to isotropic reference for dB patterns
Eiso = em.lib.EISO
plot_ff(
    ff1.ang*180/np.pi,
    [ff1.Elhcp/Eiso, ff1.Erhcp/Eiso, ff2.Elhcp/Eiso, ff2.Erhcp/Eiso],
    dB=True,
    labels=['LHCP 1','RHCP 1','LHCP 2','RHCP 2']
)

# --- 3D far-field visualization ---------------------------------------------
# Add geometry for context
model.display.add_object(helix)
model.display.add_object(airbox)

# Compute full 3D far-field (at the same frequency) and display |Erhcp|
model.display.add_surf(*data.field.find(freq=3.1e9).farfield_3d(abc_sel).surfplot('Elhcp','abs',rmax=L/2, isotropic=True, offset=(0,0,L)))

# Show interactive 3D scene
model.display.show()



# EMerge is an open source Python based FEM EM simulation module.
# Copyright (C) 2025  Robert Fennis.

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, see
# <https://www.gnu.org/licenses/>.

import numpy as np
from ...elements.nedleg2 import NedelecLegrange2
from ...mth.integrals import surface_integral


def compute_avg_power_flux(field: NedelecLegrange2, mode: np.ndarray, k0: float, ur: np.ndarray, beta: float):

    Efunc = field.interpolate_Ef(mode)
    Hfunc = field.interpolate_Hf(mode, k0, ur, beta)
    nx, ny, nz = field.mesh.normals[:,0]
    def S(x,y,z):
        Ex, Ey, Ez = Efunc(x,y,z)
        Hx, Hy, Hz = Hfunc(x,y,z)
        Sx = 1/2*np.real(Ey*np.conj(Hz) - Ez*np.conj(Hy))
        Sy = 1/2*np.real(Ez*np.conj(Hx) - Ex*np.conj(Hz))
        Sz = 1/2*np.real(Ex*np.conj(Hy) - Ey*np.conj(Hx))
        return nx*Sx + ny*Sy + nz*Sz

    Ptot = surface_integral(field.mesh.nodes, field.mesh.tris, S, None, 4)
    return Ptot

def compute_peak_power_flux(field: NedelecLegrange2, mode: np.ndarray, k0: float, ur: np.ndarray, beta: float):

    Efunc = field.interpolate_Ef(mode)
    Hfunc = field.interpolate_Hf(mode, k0, ur, beta)
    nx, ny, nz = field.mesh.normals[:,0]
    def S(x,y,z):
        Ex, Ey, Ez = Efunc(x,y,z)
        Hx, Hy, Hz = Hfunc(x,y,z)
        Sx = np.real(Ey*np.conj(Hz) - Ez*np.conj(Hy))
        Sy = np.real(Ez*np.conj(Hx) - Ex*np.conj(Hz))
        Sz = np.real(Ex*np.conj(Hy) - Ey*np.conj(Hx))
        return nx*Sx + ny*Sy + nz*Sz

    Ptot = surface_integral(field.mesh.nodes, field.mesh.tris, S, None, 4)
    return Ptot


# EMerge is an open source Python based FEM EM simulation module.
# Copyright (C) 2025  Robert Fennis.

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, see
# <https://www.gnu.org/licenses/>.

import numpy as np

# Function to animate a field oscillation with phase shift
def animate_field(X, Y, data: np.ndarray, Nsteps: int, vrange: float | None = None) -> None:
    import matplotlib.pyplot as plt
    from matplotlib.animation import FuncAnimation
    # Compute the magnitude for consistent color scaling
    data_magnitude = np.abs(data)
    real_data_max = np.max(data_magnitude)
    if vrange is not None:
        real_data_max = vrange
    real_data_min = -real_data_max

    # Set up the figure and axis
    fig, ax = plt.subplots(figsize=(8, 6))

    # Initial field with zero phase
    psi = np.exp(-1j * 2 * np.pi * 0 / Nsteps)
    real_data = np.real(data * psi)

    # Create initial contour plot
    levels = np.linspace(real_data_min, real_data_max, 100)
    contour = ax.contourf(X, Y, real_data, levels=levels, vmin=real_data_min, vmax=real_data_max)

    # Add colorbar
    cbar = fig.colorbar(contour, ax=ax, label="Field Intensity")

    # Set plot labels and title
    ax.set_title("Field Distribution")
    ax.set_xlabel("X Coordinate")
    ax.set_ylabel("Y Coordinate")
    ax.axis("equal")

    # Animation update function
    def update(frame):
        nonlocal contour
        # Remove old contours
        for c in contour.collections:
            c.remove()
        # Update the phase
        psi = np.exp(1j * 2 * np.pi * frame / Nsteps)
        real_data = np.real(data * psi)
        # Redraw contour plot
        contour = ax.contourf(X, Y, real_data, levels=levels, vmin=real_data_min, vmax=real_data_max)
        return contour.collections

    # Create animation
    anim = FuncAnimation(fig, update, frames=Nsteps, blit=False, interval=25, repeat=True)

    plt.show()

# def plot_field(dataset: Dataset2D, data: np.ndarray, cmap=EMERGE_Base) -> None:
#         """
#         Plot the field values on the 2D mesh.

#         Parameters:
#         field (np.ndarray): Field values at each vertex in the mesh.
#         """
#         # Extract vertices and triangles
#         vertices = dataset.vertices
#         triangles = (
#             dataset.triangles.T
#         )  # Transpose to match the (N, 3) format for Matplotlib

#         plt.figure(figsize=(8, 6))
#         plt.tricontourf(
#             vertices[0,:], vertices[1,:], triangles, data, levels=100, cmap=cmap
#         )
#         plt.colorbar(label="Field Intensity")
#         plt.title("Field Distribution")
#         plt.xlabel("X Coordinate")
#         plt.ylabel("Y Coordinate")
#         plt.axis("equal")
#         plt.show()

# def plot_field_tri(mesh: Mesh2D, data: np.ndarray, cmap=EMERGE_Base) -> None:
#         """
#         Plot the field values on the 2D mesh.

#         Parameters:
#         field (np.ndarray): Field values at each vertex in the mesh.
#         """
#         # Extract vertices and triangles
#         vertices = mesh.vertices
#         triangles = (
#             mesh.triangles.T
#         )  # Transpose to match the (N, 3) format for Matplotlib

#         plt.figure(figsize=(8, 6))
#         plt.tripcolor(
#             vertices[0,:], vertices[1,:], triangles, facecolors=data, cmap=cmap
#         )
#         plt.colorbar(label="Field Intensity")
#         plt.title("Field Distribution")
#         plt.xlabel("X Coordinate")
#         plt.ylabel("Y Coordinate")
#         plt.axis("equal")
#         plt.show()

# def radiation_plot(angles: np.ndarray, amplitude) -> None:
#     fig, ax = plt.subplots(subplot_kw={'projection': 'polar'})

#     amax = np.max(np.abs(amplitude))

#     xs = amplitude*np.cos(angles)
#     ys = amplitude*np.sin(angles)


#     ax.plot(angles, amplitude, color='k')
#     #ax.set_xlim([-amax, amax])
#     #ax.set_ylim([-amax, amax])
#     ax.grid(True)


#     plt.show()

# def linplot(x, ys):
#     if not isinstance(ys, (list,tuple)):
#         ys = [ys,]
#     plt.figure(figsize=(8,6))
#     for y in ys:
#         plt.plot(x,y)
#     plt.grid(True)
#     plt.show()


# def plot_s_parameters(dataset: Dataset2D, sparam_labels=None, title="S-Parameter Plot"):
#     """
#     Plots S-parameters in dB (magnitude) and phase (degrees) across frequencies.

#     Parameters:
#     - datasets: List of Dataset2D objects, each representing a single frequency.
#                 Each dataset should contain:
#                     - dataset.frequency: The frequency at which the S-parameters are measured.
#                     - dataset.Sparam: A list of complex S-parameters (S11, S12, S21, etc.) at this frequency.
#     - sparam_labels: Optional list of labels for each S-parameter (e.g., ["S11", "S12", "S21", ...]).
#                      Defaults to S11, S12, ... based on the length of the first dataset's Sparam list.
#     - title: Title of the plot.

#     Returns:
#     - A matplotlib figure with two subplots:
#         - Top subplot: Magnitude in dB for each S-parameter across frequencies.
#         - Bottom subplot: Phase in degrees for each S-parameter across frequencies.
#     """
#     # Extract frequency and S-parameter data across all datasets
#     frequencies = dataset.freqs

#     # If no labels are provided, auto-generate based on the first dataset's S-parameter count
#     if sparam_labels is None:
#         sparam_labels = [f"S{i+1}{1}" for i in range(dataset.S.shape[1])]
    
#     sublines = [[],[]]

#     S = dataset.S
#     # Plot magnitude in dB for each S-parameter
#     for idx, label in enumerate(sparam_labels):
#         magnitudes_db = 20 * np.log10(np.abs(S[:,idx,0].squeeze()))
#         sublines[0].append(Line(frequencies, magnitudes_db, name=label, color='k'))
    

#     # Plot phase in degrees for each S-parameter
#     for idx, label in enumerate(sparam_labels):
#         phases_deg = np.angle(S[:,idx,0].squeeze(), deg=True)
#         sublines[1].append(Line(frequencies, phases_deg, name=label, color='k'))
    

#     with SubFigs(2,1) as axes:
#         eplot(sublines[0], axes=axes[0], linestyle_cycle=['-','--'],xlabel='Frequency [Hz]', ylabel='Sparam [dB]')
#         eplot(sublines[1], axes=axes[1], linestyle_cycle=['-','--'],xlabel='Frequency [Hz]', ylabel='Sparam [deg]')

# def plot_mesh(
#         vertices,
#         triangles,
#         highlight_vertices: list[int] = None,
#         highlight_triangles: list[int] = None,
#     ) -> None:
#     # Extract x and y coordinates of vertices
#     x = vertices[0, :]
#     y = vertices[1, :]

#     # Create the plot
#     plt.figure(figsize=(8, 6))
#     plt.triplot(
#         x, y, triangles.T, color="black", lw=0.5
#     )  # Use triplot for 2D triangular meshes
#     plt.scatter(x, y, color="red", s=10)  # Plot vertices for reference

#     # Set plot labels and title
#     plt.xlabel("X")
#     plt.ylabel("Y")
#     plt.title("2D Mesh Plot")
#     plt.gca().set_aspect("equal", adjustable="box")  # Set equal aspect ratio

#     if highlight_triangles:
#         for index in highlight_triangles:
#             plt.fill(
#                 x[triangles[:, index]],
#                 y[triangles[:, index]],
#                 alpha=0.3,
#                 color="blue",
#             )
#     if highlight_vertices:
#         ids = np.array(highlight_vertices)
#         x = vertices[0, ids]
#         y = vertices[1, ids]
#         plt.scatter(x, y)
#     # Show the plot
#     plt.show()


import emerge as em
from emerge.plot import plot_sp
import numpy as np

""" LUMPED COMPONENT FILTER DEMO

This demo constructs a microstrip filter using discrete lumped inductors and
capacitors mounted on a PCB. We extract the characteristic impedance, place
lumped elements in the trace, mesh the structure, and compute S-parameters.
"""

# --- Unit definitions -----------------------------------------------------
mm = 0.001           # meters per millimeter
pF = 1e-12           # picofarad in farads
fF = 1e-15           # femtofarad in farads
pH = 1e-12           # picohenry in henrys
nH = 1e-9            # nanohenry in henrys

# --- Lumped element impedance functions ---------------------------------
def Lf(L):
    """Return series impedance function for inductor L."""
    return lambda f: 1j * 2 * np.pi * f * L

def Cf(C):
    """Return shunt admittance function for capacitor C."""
    return lambda f: 1 / (1j * 2 * np.pi * f * C)

# --- PCB and lumped-component parameters ---------------------------------
pack = '0603'         # package footprint for lumped components
# Create simulation and PCB layouter with substrate thickness and material
m = em.Simulation('LumpedFilter', loglevel='DEBUG')
th = 0.5         # substrate thickness (meters)
Hair = 2.0
pcb = em.geo.PCB(th, unit=mm, cs=em.GCS,
                          material=em.lib.DIEL_RO4003C, layers=2)
# Compute 50-ohm microstrip width
w0 = pcb.calc.z0(50)

# Define lengths and lumped element values
l0 = 2            # straight segment length (mm)
Lshunt = 35 * nH         # first inductor value
Cshunt = 35 * pF         # first capacitor value
Lseries = 70 * nH         # second inductor
Cseries = 18 * pF         # second capacitor

# --- Route trace with lumped elements via method chaining ----------------
# Input matching section
pcb.new(0, 0, w0, (1, 0)).store('p1').straight(3)\
    .lumped_element(Cf(Cseries), pack).straight(l0)\
    .lumped_element(Lf(Lseries), pack).straight(l0)\
    .split((0, -1)).straight(l0)\
    .lumped_element(Cf(Cshunt), pack).straight(l0 / 2, w0)\
    .via(pcb.z(1), w0 / 6, False).merge()\
    .split((0, 1)).straight(l0)\
    .lumped_element(Lf(Lshunt), pack).straight(l0 / 2, w0)\
    .via(pcb.z(1), w0 / 6, False).merge()\
    .straight(l0).lumped_element(Cf(Cseries), pack).straight(l0)\
    .lumped_element(Lf(Lseries), pack).straight(l0)\
    .straight(3).store('p2')           

# Retrieve lumped element and via objects
LEs = pcb.lumped_elements
vias = pcb.generate_vias(merge=True)
# Compile trace geometry and determine bounds
traces = pcb.compile_paths(merge=True)
pcb.determine_bounds(leftmargin=0, topmargin=1, rightmargin=0, bottommargin=1)

# --- Define modal ports and generate environment ------------------------
mp1 = pcb.modal_port(pcb.load('p1'), Hair)
mp2 = pcb.modal_port(pcb.load('p2'), Hair)
diel = pcb.generate_pcb()                   # substrate dielectric block
air = pcb.generate_air(Hair)                  # surrounding air block

# Add all geometry to simulation
m.commit_geometry()

# --- Solver and mesh settings -------------------------------------------
m.mw.set_frequency_range(0.05e9, 0.3e9, 51)       # 50–300 MHz sweep
m.mesher.set_boundary_size(traces, 0.5 * mm)
# Refine mesh around lumped component faces
for le in LEs:
    m.mesher.set_face_size(le, 0.1 * mm)
# Domain mesh refinement
m.mesher.set_domain_size(diel, 1 * mm)
m.mesher.set_domain_size(air, 3 * mm)

# Build mesh and view
m.generate_mesh()
m.view()

# --- Boundary conditions -----------------------------------------------
# Define modal (TEM) ports at input and output
p1 = m.mw.bc.ModalPort(mp1, 1, TEM=True)
p2 = m.mw.bc.ModalPort(mp2, 2, TEM=True)
# Add lumped element BCs for each element
for le in LEs:
    m.mw.bc.LumpedElement(le)
# Perfect conductor on copper traces and vias
m.mw.bc.PEC(traces)
m.mw.bc.PEC(vias.outside())

# --- Run frequency-domain simulation ------------------------------------
data = m.mw.run_sweep(parallel=True, njobs=4, frequency_groups=8)

# --- Post-processing: plot S-parameters ---------------------------------
f = data.scalar.grid.freq
S11 = data.scalar.grid.S(1, 1)
S21 = data.scalar.grid.S(2, 1)
plot_sp(f, [S11, S21], xunit='MHz', labels=['S11', 'S21'])

# --- Visualize field distribution ---------------------------------------
m.display.add_object(diel, opacity=0.1)
m.display.add_object(traces, opacity=0.1)
# Cut-plane of Ez field through substrate center
cut = data.field.find(freq=2.25e9)\
    .cutplane(0.1 * mm, z=-th/2 * mm)
m.display.add_surf(*cut.scalar('Ez', 'real'))
m.display.show()
